@@ -9,9 +9,9 @@
 # Contact email: d.lowe@rl.ac.uk
 # =============================================================================
 
-from wcsBase import WCSBase, WCSCapabilitiesReader, RereadableURL
+from wcsBase import WCSBase, WCSCapabilitiesReader
 from urllib import urlencode
-from urllib2 import urlopen, HTTPError
+from owslib.util import openURL, testXMLValue
 from owslib.etree import etree
 import os, errno
 
@@ -34,11 +34,12 @@ class WebCoverageService_1_0_0(WCSBase):
         else:
             raise KeyError, "No content named %s" % name
     
-    def __init__(self,url,xml):
+    def __init__(self,url,xml, cookies):
         self.version='1.0.0'
         self.url = url   
+        self.cookies=cookies
         # initialize from saved capability document or access the server
-        reader = WCSCapabilitiesReader(self.version)
+        reader = WCSCapabilitiesReader(self.version, self.cookies)
         if xml:
             self._capabilities = reader.readString(xml)
         else:
@@ -46,17 +47,15 @@ class WebCoverageService_1_0_0(WCSBase):
 
         #serviceIdentification metadata
         subelem=self._capabilities.find(ns('Service'))
-        self.identification=ServiceIdenfication(subelem)                               
+        self.identification=ServiceIdentification(subelem)                               
                    
         #serviceProvider metadata
-        self.provider=None
         subelem=self._capabilities.find(ns('Service/')+ns('responsibleParty'))
-        if subelem is not None:
-            self.provider=ServiceProvider(subelem)   
+        self.provider=ServiceProvider(subelem)   
         
         #serviceOperations metadata
         self.operations=[]
-        for elem in self._capabilities.find(ns('Capability/')+ns('Request')).getchildren():
+        for elem in self._capabilities.find(ns('Capability/')+ns('Request'))[:]:
             self.operations.append(OperationMetadata(elem))
           
         #serviceContents metadata
@@ -65,6 +64,12 @@ class WebCoverageService_1_0_0(WCSBase):
             cm=ContentMetadata(elem, self)
             self.contents[cm.id]=cm
         
+        #Some WCS servers (wrongly) advertise 'Content' OfferingBrief instead.
+        if self.contents=={}:
+            for elem in self._capabilities.findall(ns('ContentMetadata/')+ns('ContentOfferingBrief')): 
+                cm=ContentMetadata(elem, self)
+                self.contents[cm.id]=cm
+        
         #exceptions
         self.exceptions = [f.text for f \
                 in self._capabilities.findall('Capability/Exception/Format')]
@@ -77,7 +82,13 @@ class WebCoverageService_1_0_0(WCSBase):
             items.append((item,self.contents[item]))
         return items
     
-      
+    def __makeString(self,value):
+        #using repr unconditionally breaks things in some circumstances if a value is already a string
+        if type(value) is not str:
+            sval=repr(value)
+        else:
+            sval = value
+        return sval
   
     def getCoverage(self, identifier=None, bbox=None, time=None, format = None,  crs=None, width=None, height=None, resx=None, resy=None, resz=None,parameter=None,method='Get',**kwargs):
         """Request and return a coverage from the WCS as a file-like object
@@ -103,7 +114,7 @@ class WebCoverageService_1_0_0(WCSBase):
         request['Coverage']=identifier
         #request['identifier'] = ','.join(identifier)
         if bbox:
-            request['BBox']=','.join([str(x) for x in bbox])
+            request['BBox']=','.join([self.__makeString(x) for x in bbox])
         else:
             request['BBox']=None
         if time:
@@ -131,28 +142,12 @@ class WebCoverageService_1_0_0(WCSBase):
         data = urlencode(request)
         self.log.debug('WCS 1.0.0 DEBUG: Second part of URL: %s'%data)
         
-        try:
-            u = urlopen(base_url + data)
-        except HTTPError, e: #Some servers may set the http header to 400 if returning an OGC service exception.
-            if e.code == 400:
-                raise ServiceException, e.read()
-  
         
-        self.log.debug('WCS 1.0.0 DEBUG: GetCoverage request made: %s'%u.url)
-        self.log.debug('WCS 1.0.0 DEBUG: Headers returned: %s'%str(u.headers))
-        # check for service exceptions, and return #TODO - test this bit properly.
-        if u.info()['Content-Type'] == 'text/xml':          
-            #just in case 400 headers were not set, going to have to read the xml to see if it's an exception report.
-            #wrap the url stram in a extended StringIO object so it's re-readable
-            u=RereadableURL(u)      
-            se_xml= u.read()
-            se_tree = etree.fromstring(se_xml)
-            serviceException=se_tree.find('{http://www.opengis.net/ows}Exception')
-            if serviceException is not None:
-                raise ServiceException, \
-                str(serviceException.text).strip()
-            u.seek(0)
+        u=openURL(base_url, data, method, self.cookies)
+
         return u
+    
+
                
     def getOperationByName(self, name):
         """Return a named operation item."""
@@ -179,17 +174,15 @@ class OperationMetadata(object):
             methods.append(('Post', {'url': url}))        
         self.methods = dict(methods)
             
-class ServiceIdenfication(object):
+class ServiceIdentification(object):
     """ Abstraction for ServiceIdentification metadata """
     def __init__(self,elem):
         # properties              
+        self.type='OGC:WCS'
         self.version='1.0.0'
-        self.service = elem.find(ns('name')).text
-        try:
-            self.abstract = elem.find(ns('description')).text
-        except:
-            self.abstract=None
-        self.title = elem.find(ns('name')).text     
+        self.service = testXMLValue(elem.find(ns('name')))
+        self.abstract = testXMLValue(elem.find(ns('description')))
+        self.title = testXMLValue(elem.find(ns('name')))     
         self.keywords = [f.text for f in elem.findall(ns('keywords')+'/'+ns('keyword'))]
         #note: differs from 'rights' in interface
         self.fees=elem.find(ns('fees')).text
@@ -199,13 +192,16 @@ class ServiceProvider(object):
     """ Abstraction for WCS ResponsibleParty 
     Implements IServiceProvider"""
     def __init__(self,elem):
-        name=elem.find(ns('organisationName'))
-        if name is not None:
-            self.name=name.text
-        else:
+        #it's not uncommon for the service provider info to be missing
+        #so handle case where None is passed in
+        if elem is None:
             self.name=None
-        self.url=self.name #there is no definitive place for url  WCS, repeat organisationName
-        self.contact=ContactMetadata(elem)
+            self.url=None
+            self.contact = None
+        else:
+            self.name=testXMLValue(elem.find(ns('organisationName')))
+            self.url=self.name #there is no definitive place for url  WCS, repeat organisationName
+            self.contact=ContactMetadata(elem)
 
 class ContactMetadata(object):
     ''' implements IContactMetadata'''
@@ -257,7 +253,8 @@ class ContentMetadata(object):
         self.id=elem.find(ns('name')).text
         self.title =elem.find(ns('label')).text       
         self.keywords = [f.text for f in elem.findall(ns('keywords')+'/'+ns('keyword'))]        
-        self.boundingBoxWGS84 = None
+        self.boundingBox=None #needed for iContentMetadata harmonisation
+        self.boundingBoxWGS84 = None        
         b = elem.find(ns('lonLatEnvelope')) 
         if b is not None:
             gmlpositions=b.findall('{http://www.opengis.net/gml}pos')
@@ -267,7 +264,10 @@ class ContentMetadata(object):
                     float(lc.split()[0]),float(lc.split()[1]),
                     float(uc.split()[0]), float(uc.split()[1]),
                     )
-        
+        #others not used but needed for iContentMetadata harmonisation
+        self.styles=None
+        self.crsOptions=None
+
     #grid is either a gml:Grid or a gml:RectifiedGrid if supplied as part of the DescribeCoverage response.
     def _getGrid(self):
         if not hasattr(self, 'descCov'):
@@ -316,7 +316,6 @@ class ContentMetadata(object):
             self.descCov=self._service.getDescribeCoverage(self.id)
         return bboxes        
     boundingboxes=property(_getOtherBoundingBoxes,None)
-
     
     def _getSupportedCRSProperty(self):
         # gets supported crs info
@@ -393,4 +392,4 @@ class AxisDescription(object):
                 self.label = elem.text
             elif elem.tag == ns('values'):
                 for child in elem.getchildren():
-                    self.values.append(child.text)     
\ No newline at end of file
+                    self.values.append(child.text)     
