@@ -17,17 +17,24 @@ Currently supports only version 1.1.1 of the WMS protocol.
 
 import cgi
 from urllib import urlencode
-from urllib2 import urlopen
-from urllib2 import HTTPPasswordMgrWithDefaultRealm
-from urllib2 import HTTPBasicAuthHandler
-from urllib2 import build_opener
-from urllib2 import install_opener
-
 from etree import etree
+from .util import openURL
 
 
 class ServiceException(Exception):
-    pass
+    """WMS ServiceException
+
+    Attributes:
+        message -- short error message
+        xml  -- full xml error message from server
+    """
+
+    def __init__(self, message, xml):
+        self.message = message
+        self.xml = xml
+        
+    def __str__(self):
+        return repr(self.message)
 
 
 class CapabilitiesError(Exception):
@@ -57,31 +64,18 @@ class WebMapService(object):
         self.password = password
         self.version = version
         self._capabilities = None
-        self._open = urlopen
         
-        if self.username and self.password:
-            # Provide login information in order to use the WMS server
-            # Create an OpenerDirector with support for Basic HTTP 
-            # Authentication...
-            passman = HTTPPasswordMgrWithDefaultRealm()
-            passman.add_password(None, self.url, self.username, self.password)
-            auth_handler = HTTPBasicAuthHandler(passman)
-            opener = build_opener(auth_handler)
-            self._open = opener.open
-            reader = WMSCapabilitiesReader(
+        # Authentication handled by Reader
+        reader = WMSCapabilitiesReader(
                 self.version, url=self.url, un=self.username, pw=self.password
                 )
-            self._capabilities = reader.readString(self.url)
+        if xml:
+            #read from stored xml
+            self._capabilities = reader.readString(xml)
         else:
-            reader = WMSCapabilitiesReader(self.version)
-            if xml:
-                #read from stored xml
-                self._capabilities = reader.readString(xml)
-            else:
-                #read from non-password protected server
-                self._capabilities = reader.read(self.url)
-                
-       
+            #read from server
+            self._capabilities = reader.read(self.url)
+
         #build metadata objects
         self._buildMetadata()
 
@@ -92,7 +86,6 @@ class WebMapService(object):
                 )
             self._capabilities = ServiceMetadata(reader.read(self.url))
         return self._capabilities
-    capabilities = property(_getcapproperty, None)
 
     def _buildMetadata(self):         
         ''' set up capabilities metadata objects '''
@@ -106,25 +99,30 @@ class WebMapService(object):
             
         #serviceOperations metadata 
         self.operations=[]
-        for elem in self._capabilities.find('Capability/Request').getchildren():
+        for elem in self._capabilities.find('Capability/Request')[:]:
             self.operations.append(OperationMetadata(elem))
           
         #serviceContents metadata: our assumption is that services use a top-level 
         #layer as a metadata organizer, nothing more.
         self.contents={}
         caps = self._capabilities.find('Capability')
-        for elem in caps.findall('Layer'):
-            cm=ContentMetadata(elem)
-            self.contents[cm.id]=cm       
-            for subelem in elem.findall('Layer'):
-                subcm=ContentMetadata(subelem, cm)
-                self.contents[subcm.id]=subcm 
+        
+        #recursively gather content metadata for all layer elements.
+        #To the WebMapService.contents store only metadata of named layers.
+        def gather_layers(parent_elem, parent_metadata):
+            for index, elem in enumerate(parent_elem.findall('Layer')):
+                cm = ContentMetadata(elem, parent=parent_metadata, index=index+1)
+                if cm.id:
+                    if cm.id in self.contents:
+                        raise KeyError('Content metadata for layer "%s" already exists' % cm.id)
+                    self.contents[cm.id] = cm
+                gather_layers(elem, cm)
+        gather_layers(caps, None)
         
         #exceptions
         self.exceptions = [f.text for f \
                 in self._capabilities.findall('Capability/Exception/Format')]
             
-            
     def items(self):
         '''supports dict-like items() access'''
         items=[]
@@ -145,15 +143,16 @@ class WebMapService(object):
         if u.info().gettype() == 'application/vnd.ogc.se_xml':
             se_xml = u.read()
             se_tree = etree.fromstring(se_xml)
-            raise ServiceException, \
-                str(se_tree.find('ServiceException').text).strip()
+            err_message = str(se_tree.find('ServiceException').text).strip()
+            raise ServiceException(err_message, se_xml)
         return u
 
     def getmap(self, layers=None, styles=None, srs=None, bbox=None,
                format=None, size=None, time=None, transparent=False,
                bgcolor='#FFFFFF',
                exceptions='application/vnd.ogc.se_xml',
-               method='Get'
+               method='Get',
+               **kwargs
                ):
         """Request and return an image from the WMS as a file-like object.
         
@@ -178,6 +177,8 @@ class WebMapService(object):
             Optional. Image background color.
         method : string
             Optional. HTTP DCP method name: Get or Post.
+        **kwargs : extra arguments
+            anything else e.g. vendor specific parameters
         
         Example
         -------
@@ -211,7 +212,7 @@ class WebMapService(object):
         request['height'] = str(size[1])
         
         request['srs'] = str(srs)
-        request['bbox'] = ','.join([str(x) for x in bbox])
+        request['bbox'] = ','.join([repr(x) for x in bbox])
         request['format'] = str(format)
         request['transparent'] = str(transparent).upper()
         request['bgcolor'] = '0x' + bgcolor[1:7]
@@ -220,19 +221,27 @@ class WebMapService(object):
         if time is not None:
             request['time'] = str(time)
         
+        if kwargs:
+            for kw in kwargs:
+                request[kw]=kwargs[kw]
+
         data = urlencode(request)
-        if method == 'Post':
-            u = self._open(base_url, data=data)
-        else:
-            u = self._open(base_url + data)
+        
+        u = openURL(base_url, data, method, username = self.username, password = self.password)
 
         # check for service exceptions, and return
         if u.info()['Content-Type'] == 'application/vnd.ogc.se_xml':
             se_xml = u.read()
             se_tree = etree.fromstring(se_xml)
-            raise ServiceException, \
-                str(se_tree.find('ServiceException').text).strip()
+            err_message = unicode(se_tree.find('ServiceException').text).strip()
+            raise ServiceException(err_message, se_xml)
         return u
+        
+    def getServiceXML(self):
+        xml = None
+        if self._capabilities:
+            xml = etree.tostring(self._capabilities)
+        return xml
 
     def getfeatureinfo(self):
         raise NotImplementedError
@@ -253,10 +262,10 @@ class ServiceIdentification(object):
         self.version = version
         self.title = self._root.find('Title').text
         abstract = self._root.find('Abstract')
-	if abstract is not None:
-	        self.abstract = abstract.text
-	else:
-		self.abstract = None
+        if abstract is not None:
+            self.abstract = abstract.text
+        else:
+            self.abstract = None
         self.keywords = [f.text for f in self._root.findall('KeywordList/Keyword')]
         accessconstraints=self._root.find('AccessConstraints')
         if accessconstraints is not None:
@@ -276,10 +285,10 @@ class ServiceProvider(object):
             self.name=None
         self.url=self._root.find('OnlineResource').attrib.get('{http://www.w3.org/1999/xlink}href', '')
         #contact metadata
-	contact = self._root.find('ContactInformation')
-	## sometimes there is a contact block that is empty, so make
-	## sure there are children to parse
-	if contact is not None and contact.getchildren() != []:
+        contact = self._root.find('ContactInformation')
+        ## sometimes there is a contact block that is empty, so make
+        ## sure there are children to parse
+        if contact is not None and contact[:] != []:
             self.contact = ContactMetadata(contact)
         else:
             self.contact = None
@@ -299,100 +308,152 @@ class ServiceProvider(object):
         raise KeyError, "No operation named %s" % name
 
 class ContentMetadata:
-	"""
-	Abstraction for WMS layer metadata.
+    """
+    Abstraction for WMS layer metadata.
 
-	Implements IContentMetadata.
-	"""
-	def __init__(self, elem, parent=None):
-		self.parent = parent
-		if elem.tag != 'Layer':
-			raise ValueError('%s should be a Layer' % (elem,))
-		for key in ('Name', 'Title', 'Attribution'):
-			val = elem.find(key)
-			if val is not None:
-				setattr(self, key.lower(), val.text.strip())
-			else:
-				setattr(self, key.lower(), None)
-                self.id=self.name #conform to new interface
-		# bboxes
-		b = elem.find('BoundingBox')
-		self.boundingBox = None
-                if b is not None:
-                    try: #sometimes the SRS attribute is (wrongly) not provided
-                        srs=b.attrib['SRS']
-                    except KeyError:
-                        srs=None
-                    self.boundingBox = (
-                            float(b.attrib['minx']),
-                            float(b.attrib['miny']),
-                            float(b.attrib['maxx']),
-                            float(b.attrib['maxy']),
-                            srs,
-                    )
-		elif self.parent:
-                    if hasattr(self.parent, 'boundingBox'):
-			self.boundingBox = self.parent.boundingBox
-                    
-		b = elem.find('LatLonBoundingBox')
-		if b is not None:
-			self.boundingBoxWGS84 = (
-				float(b.attrib['minx']),
-				float(b.attrib['miny']),
-				float(b.attrib['maxx']),
-				float(b.attrib['maxy']),
-			)
-		elif self.parent:
-			self.boundingBoxWGS84 = self.parent.boundingBoxWGS84
-		else:
-			self.boundingBoxWGS84 = None
-		# crs options
-		self.crsOptions = []
-		if elem.find('SRS') is not None:
-			## some servers found in the wild use a single SRS
-			## tag containing a whitespace separated list of SRIDs
-			## instead of several SRS tags. hence the inner loop
-			for srslist in map(lambda x: x.text, elem.findall('SRS')):
-                            if srslist:
-				for srs in srslist.split():
-					self.crsOptions.append(srs)
-		elif self.parent:
-                        self.crsOptions = self.parent.crsOptions
-		else:
-			#raise ValueError('%s no SRS available!?' % (elem,))
-                        #Comment by D Lowe.
-                        #Do not raise ValueError as it is possible that a layer is purely a parent layer and does not have SRS specified. Instead set crsOptions to None
-                        self.crsOptions=None
-		# styles
-		self.styles = {}
-		for s in elem.findall('Style'):
-			name = s.find('Name')
-			title = s.find('Title')
-			if name is None or title is None:
-				raise ValueError('%s missing name or title' % (s,))
-			style = { 'title' : title.text }
-			# legend url
-			legend = s.find('LegendURL/OnlineResource')
-			if legend is not None:
-				style['legend'] = legend.attrib['{http://www.w3.org/1999/xlink}href']
-			self.styles[name.text] = style
+    Implements IContentMetadata.
+    """
+    def __init__(self, elem, parent=None, index=0):
+        if elem.tag != 'Layer':
+            raise ValueError('%s should be a Layer' % (elem,))
+        
+        self.parent = parent
+        if parent:
+            self.index = "%s.%d" % (parent.index, index)
+        else:
+            self.index = str(index)
+        
+        # title is mandatory property
+        self.title = elem.find('Title').text.strip()
+        name = elem.find('Name')
+        self.name = name.text.strip() if name is not None else None
+        
+        self.id=self.name #conform to new interface
+        # bboxes
+        b = elem.find('BoundingBox')
+        self.boundingBox = None
+        if b is not None:
+            try: #sometimes the SRS attribute is (wrongly) not provided
+                srs=b.attrib['SRS']
+            except KeyError:
+                srs=None
+            self.boundingBox = (
+                float(b.attrib['minx']),
+                float(b.attrib['miny']),
+                float(b.attrib['maxx']),
+                float(b.attrib['maxy']),
+                srs,
+                )
+        elif self.parent:
+            if hasattr(self.parent, 'boundingBox'):
+                self.boundingBox = self.parent.boundingBox
 
-		# keywords
-		self.keywords = [f.text for f in elem.findall('KeywordList/Keyword')]
+        # ScaleHint 
+ 	sh = elem.find('ScaleHint') 
+ 	self.scaleHint = None 
+ 	if sh is not None: 
+ 	    self.scaleHint = {'min': sh.attrib['min'], 'max': sh.attrib['max']} 
 
-                # timepositions - times for which data is available.
-                self.timepositions=None
-                for extent in elem.findall('Extent'):
-                    if extent.attrib.get("name").lower() =='time':
-                        self.timepositions=extent.text.split(',')
-                        break
+        attribution = elem.find('Attribution')
+        if attribution is not None:
+            self.attribution = dict()
+            title = attribution.find('Title')
+            url = attribution.find('OnlineResource')
+            logo = attribution.find('LogoURL')
+            if title is not None: 
+                self.attribution['title'] = title.text
+            if url is not None:
+                self.attribution['url'] = url.attrib['{http://www.w3.org/1999/xlink}href']
+            if logo is not None: 
+                self.attribution['logo_size'] = (int(logo.attrib['width']), int(logo.attrib['height']))
+                self.attribution['logo_url'] = logo.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href']
+
+        b = elem.find('LatLonBoundingBox')
+        if b is not None:
+            self.boundingBoxWGS84 = (
+                float(b.attrib['minx']),
+                float(b.attrib['miny']),
+                float(b.attrib['maxx']),
+                float(b.attrib['maxy']),
+            )
+        elif self.parent:
+            self.boundingBoxWGS84 = self.parent.boundingBoxWGS84
+        else:
+            self.boundingBoxWGS84 = None
+            
+        #SRS options
+        self.crsOptions = []
+            
+        #Copy any parent SRS options (they are inheritable properties)
+        if self.parent:
+            self.crsOptions = list(self.parent.crsOptions)
+
+        #Look for SRS option attached to this layer
+        if elem.find('SRS') is not None:
+            ## some servers found in the wild use a single SRS
+            ## tag containing a whitespace separated list of SRIDs
+            ## instead of several SRS tags. hence the inner loop
+            for srslist in map(lambda x: x.text, elem.findall('SRS')):
+                if srslist:
+                    for srs in srslist.split():
+                        self.crsOptions.append(srs)
+                        
+        #Get rid of duplicate entries
+        self.crsOptions = list(set(self.crsOptions))
+
+        #Set self.crsOptions to None if the layer (and parents) had no SRS options
+        if len(self.crsOptions) == 0:
+            #raise ValueError('%s no SRS available!?' % (elem,))
+            #Comment by D Lowe.
+            #Do not raise ValueError as it is possible that a layer is purely a parent layer and does not have SRS specified. Instead set crsOptions to None
+            self.crsOptions=None
+            
+        #Styles
+        self.styles = {}
+        
+        #Copy any parent styles (they are inheritable properties)
+        if self.parent:
+            self.styles = self.parent.styles.copy()
+ 
+        #Get the styles for this layer (items with the same name are replaced)
+        for s in elem.findall('Style'):
+            name = s.find('Name')
+            title = s.find('Title')
+            if name is None or title is None:
+                raise ValueError('%s missing name or title' % (s,))
+            style = { 'title' : title.text }
+            # legend url
+            legend = s.find('LegendURL/OnlineResource')
+            if legend is not None:
+                style['legend'] = legend.attrib['{http://www.w3.org/1999/xlink}href']
+            self.styles[name.text] = style
+
+        # keywords
+        self.keywords = [f.text for f in elem.findall('KeywordList/Keyword')]
+
+        # timepositions - times for which data is available.
+        self.timepositions=None
+        for extent in elem.findall('Extent'):
+            if extent.attrib.get("name").lower() =='time':
+                self.timepositions=extent.text.split(',')
+                break
+
+        # MetadataURLs
+        self.metadataUrls = []
+        for m in elem.findall('MetadataURL'):
+            metadataUrl = {
+                'type': m.attrib['type'],
+                'format': m.find('Format').text.strip(),
+                'url': m.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href']
+            }
+            self.metadataUrls.append(metadataUrl)
                 
-		self.layers = []
-		for child in elem.findall('Layer'):
-			self.layers.append(ContentMetadata(child, self))
+        self.layers = []
+        for child in elem.findall('Layer'):
+            self.layers.append(ContentMetadata(child, self))
 
-	def __str__(self):
-		return 'Layer Name: %s Title: %s' % (self.name, self.title)
+    def __str__(self):
+        return 'Layer Name: %s Title: %s' % (self.name, self.title)
 
 
 class OperationMetadata:
@@ -412,49 +473,48 @@ class OperationMetadata:
         self.methods = dict(methods)
 
 class ContactMetadata:
-	"""Abstraction for contact details advertised in GetCapabilities.
-	"""
-	def __init__(self, elem):
-		name = elem.find('ContactPersonPrimary/ContactPerson')
-		if name is not None:
-                    self.name=name.text
-                else:
-                    self.name=None
-                email = elem.find('ContactElectronicMailAddress')
-                if email is not None:
-                    self.email=email.text
-                else:
-                    self.email=None
-		self.address = self.city = self.region = None
-		self.postcode = self.country = None
+    """Abstraction for contact details advertised in GetCapabilities.
+    """
+    def __init__(self, elem):
+        name = elem.find('ContactPersonPrimary/ContactPerson')
+        if name is not None:
+            self.name=name.text
+        else:
+            self.name=None
+        email = elem.find('ContactElectronicMailAddress')
+        if email is not None:
+            self.email=email.text
+        else:
+            self.email=None
+        self.address = self.city = self.region = None
+        self.postcode = self.country = None
 
-		address = elem.find('ContactAddress')
-		if address is not None:
-			street = address.find('Address')
-			if street is not None: self.address = street.text
+        address = elem.find('ContactAddress')
+        if address is not None:
+            street = address.find('Address')
+            if street is not None: self.address = street.text
 
-			city = address.find('City')
-			if city is not None: self.city = city.text
+            city = address.find('City')
+            if city is not None: self.city = city.text
 
-			region = address.find('StateOrProvince')
-			if region is not None: self.region = region.text
+            region = address.find('StateOrProvince')
+            if region is not None: self.region = region.text
 
-			postcode = address.find('PostCode')
-			if postcode is not None: self.postcode = postcode.text
+            postcode = address.find('PostCode')
+            if postcode is not None: self.postcode = postcode.text
 
-			country = address.find('Country')
-			if country is not None: self.country = country.text
+            country = address.find('Country')
+            if country is not None: self.country = country.text
 
-		organization = elem.find('ContactPersonPrimary/ContactOrganization')
-		if organization is not None: self.organization = organization.text
-		else:self.organization = None
+        organization = elem.find('ContactPersonPrimary/ContactOrganization')
+        if organization is not None: self.organization = organization.text
+        else:self.organization = None
 
-		position = elem.find('ContactPosition')
-		if position is not None: self.position = position.text
-		else: self.position = None
+        position = elem.find('ContactPosition')
+        if position is not None: self.position = position.text
+        else: self.position = None
 
-
-        
+      
 class WMSCapabilitiesReader:
     """Read and parse capabilities document into a lxml.etree infoset
     """
@@ -466,17 +526,16 @@ class WMSCapabilitiesReader:
         self.url = url
         self.username = un
         self.password = pw
-        self._open = urlopen
 
-        if self.username and self.password:
-            # Provide login information in order to use the WMS server
-            # Create an OpenerDirector with support for Basic HTTP 
-            # Authentication...
-            passman = HTTPPasswordMgrWithDefaultRealm()
-            passman.add_password(None, self.url, self.username, self.password)
-            auth_handler = HTTPBasicAuthHandler(passman)
-            opener = build_opener(auth_handler)
-            self._open = opener.open
+        #if self.username and self.password:
+            ## Provide login information in order to use the WMS server
+            ## Create an OpenerDirector with support for Basic HTTP 
+            ## Authentication...
+            #passman = HTTPPasswordMgrWithDefaultRealm()
+            #passman.add_password(None, self.url, self.username, self.password)
+            #auth_handler = HTTPBasicAuthHandler(passman)
+            #opener = build_opener(auth_handler)
+            #self._open = opener.open
 
     def capabilities_url(self, service_url):
         """Return a capabilities url
@@ -504,8 +563,11 @@ class WMSCapabilitiesReader:
         service_url is the base url, to which is appended the service,
         version, and request parameters
         """
-        request = self.capabilities_url(service_url)
-        u = self._open(request)
+        getcaprequest = self.capabilities_url(service_url)
+
+        #now split it up again to use the generic openURL function...
+        spliturl=getcaprequest.split('?')
+        u = openURL(spliturl[0], spliturl[1], method='Get', username = self.username, password = self.password)
         return etree.fromstring(u.read())
 
     def readString(self, st):
@@ -516,27 +578,3 @@ class WMSCapabilitiesReader:
         if not isinstance(st, str):
             raise ValueError("String must be of type string, not %s" % type(st))
         return etree.fromstring(st)
-
-
-class WMSError(Exception):
-    """Base class for WMS module errors
-    """
-
-    def __init__(self, message):
-        """Initialize a WMS Error"""
-        self.message = message
-
-    def toxml(self):
-        """Serialize into a WMS Service Exception XML
-        """
-        preamble = '<?xml version="1.0" ?>'
-        report_elem = etree.Element('ServiceExceptionReport')
-        report_elem.attrib['version'] = '1.1.1'
-        # Service Exception
-        exception_elem = etree.Element('ServiceException')
-        exception_elem.text = self.message
-        report_elem.append(exception_elem)
-        return preamble + etree.tostring(report_elem)
-
-
-
